@@ -20,6 +20,9 @@
 #include "shumate-map-layer.h"
 #include "shumate-memory-cache-private.h"
 #include "shumate-tile-private.h"
+#include "shumate-symbol-event.h"
+#include "shumate-profiling-private.h"
+#include "shumate-utils-private.h"
 
 #ifdef SHUMATE_HAS_VECTOR_RENDERER
 #  include "vector/shumate-vector-symbol-container-private.h"
@@ -47,6 +50,14 @@ struct _ShumateMapLayer
 
   ShumateMemoryCache *memcache;
 
+  gint64 profile_all_tiles_filled_begin;
+  gint64 profile_all_tiles_done_begin;
+
+  guint defer_callback_id;
+  double defer_latitude_y, defer_longitude_x, defer_zoom_level;
+  gint64 defer_frame_time;
+  gboolean deferring;
+
 #ifdef SHUMATE_HAS_VECTOR_RENDERER
   ShumateVectorSymbolContainer *symbols;
 #endif
@@ -62,6 +73,14 @@ enum
 
 static GParamSpec *obj_properties[N_PROPERTIES] = { NULL, };
 
+enum
+{
+  SYMBOL_CLICKED,
+  LAST_SIGNAL
+};
+
+static guint signals[LAST_SIGNAL] = { 0, };
+
 /* This struct represents the location of a tile on the screen. It is the key
  * for the hash table tile_children which stores all visible tiles.
  *
@@ -69,60 +88,6 @@ static GParamSpec *obj_properties[N_PROPERTIES] = { NULL, };
  * here are *not* wrapped. For example, a ShumateTile at level 3 might have
  * coordinates of (7, 2) but have a TileGridPosition of (-1, 2). */
 
-typedef struct
-{
-  int x;
-  int y;
-  int zoom;
-} TileGridPosition;
-
-static void
-tile_grid_position_init (TileGridPosition *self,
-                         int x,
-                         int y,
-                         int zoom)
-{
-  self->x = x;
-  self->y = y;
-  self->zoom = zoom;
-}
-
-static TileGridPosition *
-tile_grid_position_new (int x,
-                        int y,
-                        int zoom)
-{
-  TileGridPosition *self = g_new0 (TileGridPosition, 1);
-  tile_grid_position_init (self, x, y, zoom);
-  return self;
-}
-
-static guint
-tile_grid_position_hash (gconstpointer pointer)
-{
-  const TileGridPosition *self = pointer;
-  return self->x ^ self->y ^ self->zoom;
-}
-
-static gboolean
-tile_grid_position_equal (gconstpointer a, gconstpointer b)
-{
-  const TileGridPosition *pos_a = a;
-  const TileGridPosition *pos_b = b;
-  return pos_a->x == pos_b->x && pos_a->y == pos_b->y && pos_a->zoom == pos_b->zoom;
-}
-
-static void
-tile_grid_position_free (gpointer pointer)
-{
-  TileGridPosition *self = pointer;
-  if (!self)
-    return;
-
-  g_free (self);
-}
-
-G_DEFINE_AUTOPTR_CLEANUP_FUNC (TileGridPosition, tile_grid_position_free);
 
 static int
 positive_mod (int i, int n)
@@ -134,7 +99,7 @@ typedef struct {
   ShumateMapLayer *self;
   ShumateTile *tile;
   char *source_id;
-  TileGridPosition pos;
+  ShumateGridPosition pos;
 } TileFilledData;
 
 static void
@@ -149,9 +114,9 @@ G_DEFINE_AUTOPTR_CLEANUP_FUNC (TileFilledData, tile_filled_data_free);
 
 
 static void
-add_symbols (ShumateMapLayer  *self,
-             ShumateTile      *tile,
-             TileGridPosition *pos)
+add_symbols (ShumateMapLayer     *self,
+             ShumateTile         *tile,
+             ShumateGridPosition *pos)
 {
 #ifdef SHUMATE_HAS_VECTOR_RENDERER
   GPtrArray *symbols;
@@ -203,12 +168,15 @@ on_tile_filled (GObject      *source_object,
 }
 
 static void
-add_tile (ShumateMapLayer  *self,
-          ShumateTile      *tile,
-          TileGridPosition *pos)
+add_tile (ShumateMapLayer     *self,
+          ShumateTile         *tile,
+          ShumateGridPosition *pos)
 {
   const char *source_id = shumate_map_source_get_id (self->map_source);
 
+  self->profile_all_tiles_filled_begin = SHUMATE_PROFILE_CURRENT_TIME;
+  self->profile_all_tiles_done_begin = SHUMATE_PROFILE_CURRENT_TIME;
+
   if (shumate_memory_cache_try_fill_tile (self->memcache, tile, source_id))
     {
       add_symbols (self, tile, pos);
@@ -233,9 +201,9 @@ add_tile (ShumateMapLayer  *self,
 }
 
 static void
-remove_tile (ShumateMapLayer  *self,
-             ShumateTile      *tile,
-             TileGridPosition *pos)
+remove_tile (ShumateMapLayer     *self,
+             ShumateTile         *tile,
+             ShumateGridPosition *pos)
 {
   GCancellable *cancellable = g_hash_table_lookup (self->tile_fill, tile);
   if (cancellable)
@@ -251,6 +219,72 @@ remove_tile (ShumateMapLayer  *self,
   g_signal_handlers_disconnect_by_func (tile, on_tile_notify_state, self);
 }
 
+static gboolean
+defer_tick_callback (GtkWidget     *widget,
+                     GdkFrameClock *frame_clock,
+                     gpointer       user_data)
+{
+  ShumateMapLayer *self = SHUMATE_MAP_LAYER (widget);
+  self->defer_callback_id = 0;
+
+  recompute_grid (self);
+  return G_SOURCE_REMOVE;
+}
+
+static gboolean
+should_defer (ShumateMapLayer *self)
+{
+  /* If the map is moving quickly, we may defer loading tiles until it slows back
+     down. That way, we don't waste resources loading tiles that will likely be gone
+     before they are done loading. */
+
+  ShumateViewport *viewport = shumate_layer_get_viewport (SHUMATE_LAYER (self));
+  ShumateMapSource *map_source = self->map_source;
+  double zoom_level = shumate_viewport_get_zoom_level (viewport);
+  double tile_size = shumate_map_source_get_tile_size_at_zoom (map_source, zoom_level);
+  double map_height = shumate_map_source_get_row_count (map_source, zoom_level) * tile_size;
+  double map_width = shumate_map_source_get_column_count (map_source, zoom_level) * tile_size;
+
+  double longitude_x = shumate_map_source_get_x (map_source, zoom_level, shumate_location_get_longitude (SHUMATE_LOCATION (viewport)));
+  double latitude_y = shumate_map_source_get_y (map_source, zoom_level, shumate_location_get_latitude (SHUMATE_LOCATION (viewport)));
+
+  double delta_x = self->defer_longitude_x * map_width - longitude_x;
+  double delta_y = self->defer_latitude_y * map_height - latitude_y;
+  double velocity = sqrt (delta_x * delta_x + delta_y * delta_y);
+  double width = gtk_widget_get_width (GTK_WIDGET (self));
+  double height = gtk_widget_get_height (GTK_WIDGET (self));
+  double diagonal = sqrt (width * width + height * height);
+  double zoom_velocity = self->defer_zoom_level - zoom_level;
+
+  gint64 frame_time;
+
+  if (!gtk_widget_get_realized (GTK_WIDGET (self)))
+    return FALSE;
+
+  frame_time = gdk_frame_clock_get_frame_time (gtk_widget_get_frame_clock (GTK_WIDGET (self)));
+  /* Only compare between frames, otherwise we might mistakenly think the
+     velocity is 0. */
+  if (frame_time == self->defer_frame_time)
+    return self->deferring;
+
+  if (velocity > diagonal * 0.25 || fabs (zoom_velocity) > 0.25)
+    {
+      if (self->defer_callback_id == 0)
+        self->defer_callback_id = gtk_widget_add_tick_callback (GTK_WIDGET (self), defer_tick_callback, NULL, NULL);
+
+      self->deferring = TRUE;
+    }
+  else
+    self->deferring = FALSE;
+
+  self->defer_latitude_y = latitude_y / map_height;
+  self->defer_longitude_x = longitude_x / map_width;
+  self->defer_zoom_level = shumate_viewport_get_zoom_level (viewport);
+  self->defer_frame_time = frame_time;
+
+  return self->deferring;
+}
+
 static void
 recompute_grid (ShumateMapLayer *self)
 {
@@ -273,6 +307,7 @@ recompute_grid (ShumateMapLayer *self)
   int source_columns = shumate_map_source_get_column_count (self->map_source, zoom_level);
 
   double rotation = shumate_viewport_get_rotation (viewport);
+  int n_tiles = 0;
 
   int size_x = MAX (
     abs ((int) (cos (rotation) *  width/2.0 - sin (rotation) * height/2.0)),
@@ -291,14 +326,16 @@ recompute_grid (ShumateMapLayer *self)
   int required_columns = tile_final_column - tile_initial_column;
   int required_rows = tile_final_row - tile_initial_row;
 
-  gboolean all_filled = TRUE;
+  gboolean defer = should_defer (self);
+
+  gboolean all_filled = TRUE, all_done = TRUE;
 
   /* First, remove all the tiles that aren't in bounds, or that are on the
    * wrong zoom level and haven't finished loading */
   g_hash_table_iter_init (&iter, self->tile_children);
   while (g_hash_table_iter_next (&iter, &key, &value))
     {
-      TileGridPosition *pos = key;
+      ShumateGridPosition *pos = key;
       ShumateTile *tile = value;
       float size = powf (2, zoom_level - pos->zoom);
       float x = pos->x * size;
@@ -320,28 +357,47 @@ recompute_grid (ShumateMapLayer *self)
     {
       for (int y = tile_initial_row; y < tile_initial_row + required_rows; y ++)
         {
-          g_autoptr(TileGridPosition) pos = tile_grid_position_new (x, y, zoom_level);
+          g_autoptr(ShumateGridPosition) pos = shumate_grid_position_new (x, y, zoom_level);
           ShumateTile *tile = g_hash_table_lookup (self->tile_children, pos);
 
-          if (!tile)
+          n_tiles ++;
+
+          if (!tile && !defer)
             {
               tile = shumate_tile_new_full (positive_mod (x, source_columns), positive_mod (y, source_rows), tile_size, zoom_level);
+              shumate_tile_set_scale_factor (tile, gtk_widget_get_scale_factor (GTK_WIDGET (self)));
               add_tile (self, tile, g_steal_pointer (&pos));
             }
 
-          if (shumate_tile_get_state (tile) != SHUMATE_STATE_DONE)
+          if (tile == NULL || shumate_tile_get_paintable (tile) == NULL)
             all_filled = FALSE;
+
+          if (tile == NULL || shumate_tile_get_state (tile) != SHUMATE_STATE_DONE)
+            all_done = FALSE;
         }
     }
 
+  if (all_done && self->profile_all_tiles_done_begin > 0)
+    {
+      g_autofree char *desc = g_strdup_printf ("Visible tiles done (%d)", n_tiles);
+      SHUMATE_PROFILE_COLLECT (self->profile_all_tiles_done_begin, desc, NULL);
+      self->profile_all_tiles_done_begin = 0;
+    }
+  if (all_filled && self->profile_all_tiles_filled_begin > 0)
+    {
+      g_autofree char *desc = g_strdup_printf ("Visible tiles filled (%d)", n_tiles);
+      SHUMATE_PROFILE_COLLECT (self->profile_all_tiles_filled_begin, desc, NULL);
+      self->profile_all_tiles_filled_begin = 0;
+    }
+
   /* If all the tiles on the current zoom level are filled, delete tiles on all
    * other zoom levels */
-  if (all_filled)
+  if (all_done)
     {
       g_hash_table_iter_init (&iter, self->tile_children);
       while (g_hash_table_iter_next (&iter, &key, &value))
         {
-          TileGridPosition *pos = key;
+          ShumateGridPosition *pos = key;
           ShumateTile *tile = value;
 
           if (pos->zoom != zoom_level)
@@ -462,6 +518,16 @@ shumate_map_layer_dispose (GObject *object)
   G_OBJECT_CLASS (shumate_map_layer_parent_class)->dispose (object);
 }
 
+#ifdef SHUMATE_HAS_VECTOR_RENDERER
+static void
+on_symbol_clicked (ShumateMapLayer              *self,
+                   ShumateSymbolEvent           *event,
+                   ShumateVectorSymbolContainer *container)
+{
+  g_signal_emit (self, signals[SYMBOL_CLICKED], 0, event);
+}
+#endif
+
 static void
 shumate_map_layer_constructed (GObject *object)
 {
@@ -475,6 +541,7 @@ shumate_map_layer_constructed (GObject *object)
 
 #ifdef SHUMATE_HAS_VECTOR_RENDERER
   self->symbols = shumate_vector_symbol_container_new (self->map_source, viewport);
+  g_signal_connect_object (self->symbols, "symbol-clicked", G_CALLBACK (on_symbol_clicked), self, G_CONNECT_SWAPPED);
   gtk_widget_set_parent (GTK_WIDGET (self->symbols), GTK_WIDGET (self));
 #endif
 }
@@ -606,7 +673,7 @@ shumate_map_layer_snapshot (GtkWidget *widget, GtkSnapshot *snapshot)
 
   while (g_hash_table_iter_next (&iter, &key, &value))
     {
-      TileGridPosition *pos = key;
+      ShumateGridPosition *pos = key;
       ShumateTile *tile = value;
       GdkPaintable *paintable = shumate_tile_get_paintable (tile);
       double size = tile_size * pow (2, zoom_level - pos->zoom);
@@ -670,12 +737,31 @@ shumate_map_layer_class_init (ShumateMapLayerClass *klass)
   g_object_class_install_properties (object_class,
                                      N_PROPERTIES,
                                      obj_properties);
+
+  /**
+   * ShumateMapLayer::symbol-clicked:
+   * @self: the [class@MapLayer] emitting the signal
+   * @event: a [class@SymbolEvent] with details about the clicked symbol.
+   *
+   * Emitted when a symbol in the map layer is clicked.
+   *
+   * Since: 1.1
+   */
+  signals[SYMBOL_CLICKED] =
+    g_signal_new ("symbol-clicked",
+                  G_OBJECT_CLASS_TYPE (object_class),
+                  G_SIGNAL_RUN_LAST,
+                  0, NULL, NULL,
+                  g_cclosure_marshal_VOID__OBJECT,
+                  G_TYPE_NONE,
+                  1,
+                  SHUMATE_TYPE_SYMBOL_EVENT);
 }
 
 static void
 shumate_map_layer_init (ShumateMapLayer *self)
 {
-  self->tile_children = g_hash_table_new_full (tile_grid_position_hash, tile_grid_position_equal, tile_grid_position_free, g_object_unref);
+  self->tile_children = g_hash_table_new_full (shumate_grid_position_hash, shumate_grid_position_equal, shumate_grid_position_free, g_object_unref);
   self->tile_fill = g_hash_table_new_full (g_direct_hash, g_direct_equal, g_object_unref, g_object_unref);
   self->memcache = shumate_memory_cache_new_full (100);
 }
